@@ -2,14 +2,18 @@
 # Licensed under the MIT License
 import io
 import json
+import socket
+import sys
 import uuid
 from copy import copy
 from datetime import timedelta
 from enum import Enum, unique
-from typing import Union, Callable
+from typing import TYPE_CHECKING, Union, Callable, Optional, Any, NoReturn
 
 import requests
+from requests import Response
 from requests.adapters import HTTPAdapter
+from urllib3.connection import HTTPConnection
 
 from ._version import VERSION
 from .data_format import DataFormat
@@ -17,6 +21,10 @@ from .exceptions import KustoServiceError
 from .response import KustoResponseDataSetV1, KustoResponseDataSetV2, KustoResponseDataSet
 from .security import _AadHelper
 
+if TYPE_CHECKING:
+    import aiohttp
+    import aiohttp.web_exceptions
+
 
 class KustoConnectionStringBuilder:
     """
@@ -47,6 +55,9 @@ class KustoConnectionStringBuilder:
         msi_auth = "MSI Authentication"
         msi_params = "MSI Params"
         az_cli = "AZ CLI"
+        interactive_login = "Interactive Login"
+        login_hint = "Login Hint"
+        domain_hint = "Domain Hint"
 
         @classmethod
         def parse(cls, key: str) -> "ValidKeywords":
@@ -100,6 +111,8 @@ class KustoConnectionStringBuilder:
                 self.authority_id,
                 self.application_token,
                 self.user_token,
+                self.login_hint,
+                self.domain_hint,
             ]
 
         def is_dict_type(self) -> bool:
@@ -107,7 +120,7 @@ class KustoConnectionStringBuilder:
 
         def is_bool_type(self) -> bool:
             """States whether a word is of type bool or not."""
-            return self in [self.aad_federated_security, self.msi_auth, self.az_cli]
+            return self in [self.aad_federated_security, self.msi_auth, self.az_cli, self.interactive_login]
 
     def __init__(self, connection_string: str):
         """
@@ -326,7 +339,7 @@ class KustoConnectionStringBuilder:
     def with_aad_managed_service_identity_authentication(
         cls, connection_string: str, client_id: str = None, object_id: str = None, msi_res_id: str = None, timeout: int = None
     ) -> "KustoConnectionStringBuilder":
-        """"
+        """
         Creates a KustoConnection string builder that will authenticate with AAD application, using
         an application token obtained from a Microsoft Service Identity endpoint. An optional user
         assigned application ID can be added to the token.
@@ -383,10 +396,26 @@ class KustoConnectionStringBuilder:
         assert callable(token_provider)
 
         kcsb = cls(connection_string)
+        kcsb[kcsb.ValidKeywords.aad_federated_security] = True
         kcsb._token_provider = token_provider
 
         return kcsb
 
+    @classmethod
+    def with_interactive_login(
+        cls, connection_string: str, login_hint: Optional[str] = None, domain_hint: Optional[str] = None
+    ) -> "KustoConnectionStringBuilder":
+        kcsb = cls(connection_string)
+        kcsb[kcsb.ValidKeywords.interactive_login] = True
+        kcsb[kcsb.ValidKeywords.aad_federated_security] = True
+        if login_hint is not None:
+            kcsb[kcsb.ValidKeywords.login_hint] = login_hint
+
+        if domain_hint is not None:
+            kcsb[kcsb.ValidKeywords.domain_hint] = domain_hint
+
+        return kcsb
+
     @property
     def data_source(self) -> str:
         """The URI specifying the Kusto service endpoint.
@@ -488,6 +517,19 @@ class KustoConnectionStringBuilder:
     def token_provider(self):
         return self._token_provider
 
+    @property
+    def interactive_login(self) -> bool:
+        val = self._internal_dict.get(self.ValidKeywords.interactive_login)
+        return val is not None and val
+
+    @property
+    def login_hint(self) -> Optional[str]:
+        return self._internal_dict.get(self.ValidKeywords.login_hint)
+
+    @property
+    def domain_hint(self) -> Optional[str]:
+        return self._internal_dict.get(self.ValidKeywords.domain_hint)
+
     def __str__(self):
         dict_copy = self._internal_dict.copy()
         for key in dict_copy:
@@ -514,6 +556,7 @@ class ClientRequestProperties:
 
     results_defer_partial_query_failures_option_name = "deferpartialqueryfailures"
     request_timeout_option_name = "servertimeout"
+    no_request_timeout_option_name = "norequesttimeout"
 
     def __init__(self):
         self._options = {}
@@ -553,7 +596,109 @@ class ClientRequestProperties:
         return json.dumps({"Options": self._options, "Parameters": self._parameters}, default=str)
 
 
-class KustoClient:
+class ExecuteRequestParams:
+    def __init__(self, database: str, payload: io.IOBase, properties: ClientRequestProperties, query: str, timeout: timedelta, request_headers: dict):
+        request_headers = copy(request_headers)
+        json_payload = None
+        if not payload:
+            json_payload = {"db": database, "csl": query}
+            if properties:
+                json_payload["properties"] = properties.to_json()
+
+            client_request_id_prefix = "KPC.execute;"
+            request_headers["Content-Type"] = "application/json; charset=utf-8"
+        else:
+            if properties:
+                request_headers.update(json.loads(properties.to_json())["Options"])
+
+            client_request_id_prefix = "KPC.execute_streaming_ingest;"
+            request_headers["Content-Encoding"] = "gzip"
+        request_headers["x-ms-client-request-id"] = client_request_id_prefix + str(uuid.uuid4())
+        if properties is not None:
+            if properties.client_request_id is not None:
+                request_headers["x-ms-client-request-id"] = properties.client_request_id
+            if properties.application is not None:
+                request_headers["x-ms-app"] = properties.application
+            if properties.user is not None:
+                request_headers["x-ms-user"] = properties.user
+            if properties.get_option(ClientRequestProperties.no_request_timeout_option_name, False):
+                timeout = KustoClient._mgmt_default_timeout
+            else:
+                timeout = properties.get_option(ClientRequestProperties.request_timeout_option_name, timeout)
+
+        timeout = (timeout or KustoClient._mgmt_default_timeout) + KustoClient._client_server_delta
+
+        self.json_payload = json_payload
+        self.request_headers = request_headers
+        self.timeout = timeout
+
+
+class HTTPAdapterWithSocketOptions(requests.adapters.HTTPAdapter):
+    def __init__(self, *args, **kwargs):
+        self.socket_options = kwargs.pop("socket_options", None)
+        self.pool_maxsize = kwargs.pop("pool_maxsize", None)
+        self.max_retries = kwargs.pop("max_retries", None)
+        super(HTTPAdapterWithSocketOptions, self).__init__(*args, **kwargs)
+
+    def init_poolmanager(self, *args, **kwargs):
+        if self.socket_options is not None:
+            kwargs["socket_options"] = self.socket_options
+        super(HTTPAdapterWithSocketOptions, self).init_poolmanager(*args, **kwargs)
+
+
+class _KustoClientBase:
+    _mgmt_default_timeout = timedelta(hours=1, seconds=30)
+    _query_default_timeout = timedelta(minutes=4, seconds=30)
+    _streaming_ingest_default_timeout = timedelta(minutes=10)
+
+    def __init__(self, kcsb: Union[KustoConnectionStringBuilder, str]):
+        self._kcsb = kcsb
+        if not isinstance(kcsb, KustoConnectionStringBuilder):
+            self._kcsb = KustoConnectionStringBuilder(kcsb)
+        self._kusto_cluster = self._kcsb.data_source
+
+        # Create a session object for connection pooling
+        self._mgmt_endpoint = "{0}/v1/rest/mgmt".format(self._kusto_cluster)
+        self._query_endpoint = "{0}/v2/rest/query".format(self._kusto_cluster)
+        self._streaming_ingest_endpoint = "{0}/v1/rest/ingest/".format(self._kusto_cluster)
+        self._request_headers = {"Accept": "application/json", "Accept-Encoding": "gzip,deflate", "x-ms-client-version": "Kusto.Python.Client:" + VERSION}
+
+    @staticmethod
+    def _kusto_parse_by_endpoint(endpoint: str, response_json: Any) -> KustoResponseDataSet:
+        if endpoint.endswith("v2/rest/query"):
+            return KustoResponseDataSetV2(response_json)
+        return KustoResponseDataSetV1(response_json)
+
+    @staticmethod
+    def _handle_http_error(
+        exception: Exception,
+        endpoint: Optional[str],
+        payload: Optional[io.IOBase],
+        response: "Union[Response, aiohttp.ClientResponse]",
+        response_json: Any,
+        response_text: Optional[str],
+    ) -> NoReturn:
+        if response.status_code == 404:
+            if payload:
+                raise KustoServiceError("The ingestion endpoint does not exist. Please enable streaming ingestion on your cluster.", response) from exception
+
+            raise KustoServiceError("The requested endpoint '{}' does not exist.".format(endpoint), response) from exception
+
+        if payload:
+            raise KustoServiceError(
+                "An error occurred while trying to ingest: Status: {0.status_code}, Reason: {0.reason}, Text: {1}.".format(response, response_text), response
+            ) from exception
+
+        if response_json:
+            raise KustoServiceError([response_json], response) from exception
+
+        if response_text:
+            raise KustoServiceError(response_text, response) from exception
+
+        raise KustoServiceError("Server error response contains no data.", response) from exception
+
+
+class KustoClient(_KustoClientBase):
     """
     Kusto client for Python.
     The client is a wrapper around the Kusto REST API.
@@ -564,9 +709,10 @@ class KustoClient:
     `execute_mgmt`: executes a KQL control command against the Kusto service.
     """
 
-    _mgmt_default_timeout = timedelta(hours=1, seconds=30)
-    _query_default_timeout = timedelta(minutes=4, seconds=30)
+    _mgmt_default_timeout = timedelta(hours=1)
+    _query_default_timeout = timedelta(minutes=4)
     _streaming_ingest_default_timeout = timedelta(minutes=10)
+    _client_server_delta = timedelta(seconds=30)
 
     # The maximum amount of connections to be able to operate in parallel
     _max_pool_size = 100
@@ -577,20 +723,69 @@ class KustoClient:
         :param kcsb: The connection string to initialize KustoClient.
         :type kcsb: azure.kusto.data.KustoConnectionStringBuilder or str
         """
-        if not isinstance(kcsb, KustoConnectionStringBuilder):
-            kcsb = KustoConnectionStringBuilder(kcsb)
-        kusto_cluster = kcsb.data_source
+        super().__init__(kcsb)
 
         # Create a session object for connection pooling
         self._session = requests.Session()
-        self._session.mount("http://", HTTPAdapter(pool_maxsize=self._max_pool_size))
-        self._session.mount("https://", HTTPAdapter(pool_maxsize=self._max_pool_size))
-        self._mgmt_endpoint = "{0}/v1/rest/mgmt".format(kusto_cluster)
-        self._query_endpoint = "{0}/v2/rest/query".format(kusto_cluster)
-        self._streaming_ingest_endpoint = "{0}/v1/rest/ingest/".format(kusto_cluster)
+        adapter = HTTPAdapterWithSocketOptions(
+            socket_options=(HTTPConnection.default_socket_options or []) + self.compose_socket_options(), pool_maxsize=self._max_pool_size
+        )
+        self._session.mount("http://", adapter)
+        self._session.mount("https://", adapter)
+
         # notice that in this context, federated actually just stands for add auth, not aad federated auth (legacy code)
-        self._auth_provider = _AadHelper(kcsb) if kcsb.aad_federated_security else None
-        self._request_headers = {"Accept": "application/json", "Accept-Encoding": "gzip,deflate", "x-ms-client-version": "Kusto.Python.Client:" + VERSION}
+        self._auth_provider = _AadHelper(self._kcsb) if self._kcsb.aad_federated_security else None
+
+    def set_http_retries(self, max_retries):
+        """
+        Set the number of HTTP retries to attempt
+        """
+        adapter = HTTPAdapterWithSocketOptions(
+            socket_options=(HTTPConnection.default_socket_options or []) + self.compose_socket_options(),
+            pool_maxsize=self._max_pool_size,
+            max_retries=max_retries,
+        )
+        self._session.mount("http://", adapter)
+        self._session.mount("https://", adapter)
+
+    @staticmethod
+    def compose_socket_options():
+        # Sends TCP Keep-Alive after MAX_IDLE_SECONDS seconds of idleness, once every INTERVAL_SECONDS seconds, and closes the connection after MAX_FAILED_KEEPALIVES failed pings (e.g. 20 => 1:00:30)
+        MAX_IDLE_SECONDS = 30
+        INTERVAL_SECONDS = 180  # Corresponds to Azure Load Balancer Service 4 minute timeout, with 1 minute of slack
+        MAX_FAILED_KEEPALIVES = 20
+
+        if (
+            sys.platform == "linux"
+            and hasattr(socket, "SOL_SOCKET")
+            and hasattr(socket, "SO_KEEPALIVE")
+            and hasattr(socket, "TCP_KEEPIDLE")
+            and hasattr(socket, "TCP_KEEPINTVL")
+            and hasattr(socket, "TCP_KEEPCNT")
+        ):
+            return [
+                (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
+                (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, MAX_IDLE_SECONDS),
+                (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, INTERVAL_SECONDS),
+                (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, MAX_FAILED_KEEPALIVES),
+            ]
+        elif (
+            sys.platform == "win32"
+            and hasattr(socket, "SOL_SOCKET")
+            and hasattr(socket, "SO_KEEPALIVE")
+            and hasattr(socket, "TCP_KEEPIDLE")
+            and hasattr(socket, "TCP_KEEPCNT")
+        ):
+            return [
+                (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
+                (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, MAX_IDLE_SECONDS),
+                (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, MAX_FAILED_KEEPALIVES),
+            ]
+        elif sys.platform == "darwin" and hasattr(socket, "SOL_SOCKET") and hasattr(socket, "SO_KEEPALIVE") and hasattr(socket, "IPPROTO_TCP"):
+            TCP_KEEPALIVE = 0x10
+            return [(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1), (socket.IPPROTO_TCP, TCP_KEEPALIVE, INTERVAL_SECONDS)]
+        else:
+            return []
 
     def execute(self, database: str, query: str, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
         """
@@ -616,7 +811,7 @@ class KustoClient:
         :return: Kusto response data set.
         :rtype: azure.kusto.data.response.KustoResponseDataSet
         """
-        return self._execute(self._query_endpoint, database, query, None, KustoClient._query_default_timeout, properties)
+        return self._execute(self._query_endpoint, database, query, None, self._query_default_timeout, properties)
 
     def execute_mgmt(self, database: str, query: str, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
         """
@@ -628,7 +823,7 @@ class KustoClient:
         :return: Kusto response data set.
         :rtype: azure.kusto.data.response.KustoResponseDataSet
         """
-        return self._execute(self._mgmt_endpoint, database, query, None, KustoClient._mgmt_default_timeout, properties)
+        return self._execute(self._mgmt_endpoint, database, query, None, self._mgmt_default_timeout, properties)
 
     def execute_streaming_ingest(
         self,
@@ -656,52 +851,27 @@ class KustoClient:
         if mapping_name is not None:
             endpoint = endpoint + "&mappingName=" + mapping_name
 
-        self._execute(endpoint, database, None, stream, KustoClient._streaming_ingest_default_timeout, properties)
+        self._execute(endpoint, database, None, stream, self._streaming_ingest_default_timeout, properties)
 
-    def _execute(self, endpoint: str, database: str, query: str, payload: io.IOBase, timeout: timedelta, properties: ClientRequestProperties = None):
+    def _execute(
+        self, endpoint: str, database: str, query: Optional[str], payload: Optional[io.IOBase], timeout: timedelta, properties: ClientRequestProperties = None
+    ):
         """Executes given query against this client"""
-        request_headers = copy(self._request_headers)
-        json_payload = None
-        if not payload:
-            json_payload = {"db": database, "csl": query}
-            if properties:
-                json_payload["properties"] = properties.to_json()
-
-            client_request_id_prefix = "KPC.execute;"
-            request_headers["Content-Type"] = "application/json; charset=utf-8"
-        else:
-            if properties:
-                request_headers.update(json.loads(properties.to_json())["Options"])
-
-            client_request_id_prefix = "KPC.execute_streaming_ingest;"
-            request_headers["Content-Encoding"] = "gzip"
-
-        request_headers["x-ms-client-request-id"] = client_request_id_prefix + str(uuid.uuid4())
-
-        if properties is not None:
-            if properties.client_request_id is not None:
-                request_headers["x-ms-client-request-id"] = properties.client_request_id
-            if properties.application is not None:
-                request_headers["x-ms-app"] = properties.application
-            if properties.user is not None:
-                request_headers["x-ms-user"] = properties.user
+        request_params = ExecuteRequestParams(database, payload, properties, query, timeout, self._request_headers)
+        json_payload = request_params.json_payload
+        request_headers = request_params.request_headers
+        timeout = request_params.timeout
 
         if self._auth_provider:
             request_headers["Authorization"] = self._auth_provider.acquire_authorization_header()
 
-        if properties:
-            timeout = properties.get_option(ClientRequestProperties.request_timeout_option_name, timeout)
-
         response = self._session.post(endpoint, headers=request_headers, data=payload, json=json_payload, timeout=timeout.seconds)
 
-        if response.status_code == 200:
-            if endpoint.endswith("v2/rest/query"):
-                return KustoResponseDataSetV2(response.json())
-            return KustoResponseDataSetV1(response.json())
+        response_json = None
+        try:
+            response_json = response.json()
+            response.raise_for_status()
+        except Exception as e:
+            self._handle_http_error(e, endpoint, payload, response, response_json, response.text)
 
-        if payload:
-            raise KustoServiceError(
-                "An error occurred while trying to ingest: Status: {0.status_code}, Reason: {0.reason}, Text: {0.text}".format(response), response
-            )
-
-        raise KustoServiceError([response.json()], response)
+        return self._kusto_parse_by_endpoint(endpoint, response_json)
