@@ -1,8 +1,15 @@
 # Copyright (c) Microsoft Corporation.
 # Licensed under the MIT License
-from typing import List, Union
+from typing import List, Union, TYPE_CHECKING
 import requests
 
+if TYPE_CHECKING:
+    try:
+        from aiohttp import ClientResponse
+    except ImportError:
+        # No aio installed, ignore
+        pass
+
 
 class KustoError(Exception):
     """Base class for all exceptions raised by the Kusto Python Client Libraries."""
@@ -11,7 +18,7 @@ class KustoError(Exception):
 class KustoServiceError(KustoError):
     """Raised when the Kusto service was unable to process a request."""
 
-    def __init__(self, messages: Union[str, List[dict]], http_response: requests.Response, kusto_response=None):
+    def __init__(self, messages: Union[str, List[dict]], http_response: "Union[requests.Response, ClientResponse, None]" = None, kusto_response=None):
         super().__init__(messages)
         self.http_response = http_response
         self.kusto_response = kusto_response
@@ -22,7 +29,10 @@ class KustoServiceError(KustoError):
 
     def is_semantic_error(self):
         """Checks if a response is a semantic error."""
-        return "Semantic error:" in self.http_response.text
+        try:
+            return "Semantic error:" in self.http_response.text
+        except AttributeError:
+            return False
 
     def has_partial_results(self):
         """Checks if a response exists."""
@@ -46,8 +56,8 @@ class KustoAuthenticationError(KustoClientError):
         self.exception = exception
         if "authority" in kwargs:
             self.authority = kwargs["authority"]
-        if "resource" in kwargs:
-            self.kusto_cluster = kwargs["resource"]
+        if "kusto_uri" in kwargs:
+            self.kusto_cluster = kwargs["kusto_uri"]
         self.kwargs = kwargs
 
     def __str__(self):
@@ -55,3 +65,10 @@ class KustoAuthenticationError(KustoClientError):
 
     def __repr__(self):
         return "KustoAuthenticationError('{}', '{}', '{}')".format(self.authentication_method, repr(self.exception), self.kwargs)
+
+
+class KustoAioSyntaxError(SyntaxError):
+    """Raised when trying to use aio syntax without installing the needed modules"""
+
+    def __init__(self):
+        super().__init__("Aio modules not installed, run 'pip install azure-kusto-data[aio]' to leverage aio capabilities")
